# ANTS-Hub - Copyright (C) Entr'ouvert

import collections
import datetime
import functools
import secrets
import zoneinfo

from django.http import JsonResponse

from ants_hub.data.models import Config, Lieu, Plage, RendezVous, TypeDeRdv

ANTS_TIMEZONE = zoneinfo.ZoneInfo('Europe/Paris')


def format_date_ants(dt):
    '''Format date as ANTS requests it, i.e. Paris local time with 'Z' suffix :/'''
    return dt.astimezone(ANTS_TIMEZONE).isoformat().split('+')[0] + 'Z'


def authenticate(func):
    @functools.wraps(func)
    def wrapper(request, *args, **kwargs):
        if not request.user.is_authenticated:
            header = request.headers.get('X-Hub-Rdv-Auth-Token', '')
            if not header:
                return JsonResponse("Missing X-HUB-RDV-AUTH-TOKEN header", status=401, safe=False)
            auth_token = Config.get(Config.REQUEST_FROM_ANTS_AUTH_TOKEN)
            if not auth_token:
                return JsonResponse("X-HUB-RDV-AUTH-TOKEN not configured", status=401, safe=False)
            if not secrets.compare_digest(header, auth_token):
                return JsonResponse("X-HUB-RDV-AUTH-TOKEN header invalid", status=401, safe=False)
        return func(request, *args, **kwargs)

    return wrapper


@authenticate
def get_managed_meeting_points(request):
    lieux = []
    for lieu in Lieu.objects.select_related('collectivite'):
        website = lieu.url or lieu.collectivite.url
        city_logo = lieu.logo_url or lieu.collectivite.logo_url
        d = {
            'id': str(lieu.id),
            'name': lieu.nom,
            'longitude': lieu.longitude,
            'latitude': lieu.latitude,
            'public_entry_address': lieu.numero_rue,
            'zip_code': lieu.code_postal,
            'city_name': lieu.ville,
            'website': website,
            'city_logo': city_logo,
        }
        lieux.append(d)
    return JsonResponse(lieux, safe=False)


def available_time_slots_parse_qs(request):
    meeting_point_ids = request.GET.getlist('meeting_point_ids')
    if not meeting_point_ids:
        raise ValueError('missing meeting_point_ids')
    try:
        meeting_point_ids = list(map(int, map(str.strip, meeting_point_ids)))
    except ValueError:
        raise ValueError('invalid meeting_point_ids')
    start_date = request.GET.get('start_date', '').strip()
    if not start_date:
        raise ValueError('missing start_date')
    try:
        start_date = datetime.date.fromisoformat(start_date)
    except ValueError:
        start_date = datetime.date.today()
    if start_date < datetime.date.today():
        start_date = datetime.date.today()
    end_date = request.GET.get('end_date', '').strip()
    if not end_date:
        raise ValueError('missing end_date')
    try:
        end_date = datetime.date.fromisoformat(end_date)
    except ValueError:
        end_date = start_date + datetime.timedelta(days=7)
    if end_date < start_date:
        end_date = start_date + datetime.timedelta(days=7)
    end_date = min(end_date, start_date + datetime.timedelta(days=180))
    reason = request.GET.get('reason', 'CNI').strip()
    try:
        reason = TypeDeRdv.from_ants_name(reason)
    except KeyError:
        reason = TypeDeRdv.CNI
    try:
        documents_number = min(int(request.GET.get('documents_number', '').strip()), 4)
    except (ValueError, TypeError):
        documents_number = 1
    return meeting_point_ids, start_date, end_date, reason, documents_number


@authenticate
def available_time_slots(request):
    try:
        meeting_point_ids, start_date, end_date, reason, documents_number = available_time_slots_parse_qs(
            request
        )
    except ValueError as e:
        return JsonResponse(
            {
                'detail': [
                    {
                        'loc': 'query-string',
                        'msg': str(e),
                        'type': 'bad request',
                    }
                ]
            },
            status=422,
        )
    types_de_rdv = [reason]
    qs = Plage.objects.filter(lieu__id__in=meeting_point_ids)
    qs = qs.filter(type_de_rdv__in=types_de_rdv)
    qs = qs.filter(date__gte=start_date, date__lte=end_date)
    qs = qs.filter(personnes=documents_number)
    qs = qs.select_related('lieu', 'lieu__collectivite')

    slots_by_lieu = collections.defaultdict(list)
    date_by_lieu = collections.defaultdict(set)
    for plage in qs.order_by('lieu', 'date', 'personnes'):
        lieu_id = str(plage.lieu_id)
        for date, url in plage.available_time_slots():
            callback_url = request.build_absolute_uri(url)
            if date not in date_by_lieu[plage.lieu_id]:
                # prevent the same datetime to be reported multiple times
                date_by_lieu[plage.lieu_id].add(date)
                slots_by_lieu[lieu_id].append(
                    {
                        'datetime': format_date_ants(date),
                        'callback_url': callback_url,
                    }
                )
    return JsonResponse(slots_by_lieu)


@authenticate
def search_application_ids(request):
    application_ids = map(str.strip, request.GET.getlist('application_ids'))
    application_ids = filter(None, application_ids)
    application_ids = list(filter(lambda x: len(x) < 64, application_ids))
    if not application_ids:
        return JsonResponse(
            {
                'detail': [
                    {
                        'loc': 'query-string',
                        'msg': 'missing application_ids',
                        'type': 'bad request',
                    }
                ]
            },
            status=422,
        )
    rdv_by_identifiant_predemande = {}
    qs = RendezVous.objects.filter(identifiant_predemande__in=application_ids)
    qs = qs.select_related('lieu', 'lieu__collectivite')
    for rdv in qs:
        rdv_by_identifiant_predemande.setdefault(rdv.identifiant_predemande, []).append(
            {
                'meeting_point': rdv.lieu.nom,
                'datetime': format_date_ants(rdv.date),
                'management_url': request.build_absolute_uri(rdv.make_gestion_url()),
                'cancel_url': request.build_absolute_uri(rdv.make_annulation_url()),
            }
        )
    return JsonResponse(rdv_by_identifiant_predemande)
